#include <cstdlib>
#include <iostream>
#include <map>
#include <set>
#include <string>

#include <sys/inotify.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <dirent.h>
#include <assert.h>

using namespace std;


/* The user playlists are in folders under this path in the form
 *
 * <username>/playlists/xy.m3u
 */
#define USER_PLAYLIST_FOLDER "/usr/syno/etc/preference"

/* The path to the sqeezebox servers playlist folder */
#define SQEEZEBOX_PLAYLISTS_FOLDER "/volume1/playlists"

/* The path to awk */
#define AWK "/usr/bin/awk"

/* The path to the awk script for converting the playlist formats */
#define CONVERTER_SCRIPT "/usr/local/audiostation-playlistmonitor/bin/playlist-as2squeeze.awk"


/* Debug stuff */
bool debug = false;
#define DOUT( args ) { if ( debug ) { cout << args ; } }

/* Helper macro for clumsy C file interface */
#define ITERATE_FOLDER( folder, dp, code ) \
{ \
  struct dirent * dp ; \
  DIR *dir = opendir( folder ); \
  while( ( dp = readdir( dir ) ) != NULL ) { \
    code ; \
  } \
  closedir( dir ); \
}

/* Helper function for clumsp C file interface */
bool isFile( string path )
{
  struct stat buf;
  int res = stat( path.c_str(), &buf );
  return ( res == 0 ) && S_ISREG( buf.st_mode );
}

/* Helper function for clumsp C file interface */
bool isFolder( string path )
{
  struct stat buf;
  int res = stat( path.c_str(), &buf );
  return ( res == 0 ) && S_ISDIR( buf.st_mode );
}


/* Terminates with an error message & code */
void terminate_program( int exitcode, string message )
{
  cerr << message << endl;
  exit( exitcode );
}

/* Returns the path to the sqeezebox server playlist corresponding to the given audiostation playlist */
string getSqueezeboxPlaylistPath( string audiostationPlaylistPath )
{
  size_t basenameBegin = audiostationPlaylistPath.rfind( '/' ) + 1;
  assert( basenameBegin > 0 && "Illegal argument!" );
  string result = string( SQEEZEBOX_PLAYLISTS_FOLDER ) + string( "/" ) +
      audiostationPlaylistPath.substr( basenameBegin,
          audiostationPlaylistPath.length() - basenameBegin );
  DOUT( "result for " << audiostationPlaylistPath << " is " << result << endl );
  return result;
}

/* Handles removed playlists */
void handlePlaylistRemoval( string path )
{
  unlink( getSqueezeboxPlaylistPath( path ).c_str() );
}

/* Handles created/modified playlists */
void handlePlaylistUpdate( string path )
{
  string command = string( AWK ) + string( " -f " ) + string( CONVERTER_SCRIPT ) +
      string( " '" ) + path + string( "' > '" ) + getSqueezeboxPlaylistPath( path ) +
      string( "'" );
  DOUT( "Issuing command " << command << endl );
  system( command.c_str() );
}


/**
 * Watches for changes in audiostation user playlists and carries them over to
 * the SqeezeBox Server playlists.
 */
int main( int argc, char **argv )
{
  if ( argc > 1 && string( argv[1] ) == "-d" ) {
    debug = true;
  }

  // Init notification facility
  int fd = inotify_init ();
  if (fd < 0)
    terminate_program( 5, "inotify_init error" );

  // Init file watching
  string userFolder( USER_PLAYLIST_FOLDER );
  map<int, string> watchedFolders;
  // Iterate over all user folders
  ITERATE_FOLDER( userFolder.c_str(), userFolderEntry,

    // Check for the existence of a "playlists" folder
    string playlistFolder = userFolder + string( "/" ) +
      string( userFolderEntry->d_name ) + string( "/playlists" );

    if ( isFolder( playlistFolder ) ) {

      // Watch the playlist folder
      int wd = inotify_add_watch (fd, playlistFolder.c_str(),
          IN_MODIFY | IN_CREATE | IN_DELETE);
      if (wd < 0)
        terminate_program( 3, "inotify_add_watch error" );

      watchedFolders[ wd ] = playlistFolder.c_str();
      DOUT( "Added " << playlistFolder << " to watch list" << endl );
    }
  );

  // Begin watching
  DOUT( "Begin watching files" << endl );
  while ( 1 ) {
    // Reserve space for 1024 file events
    const int buf_len = 1024 * ( sizeof( struct inotify_event) + 16 );
    char buf[buf_len];
    // Listen to the events
    int len = read( fd, buf, buf_len );

    // List of files that we need to renew. We first gather them here, since
    // otherwise there will be too many events for a single write to the file
    set<string> filesToUpdate;

    if (len <=0 ) {
      terminate_program( 4, "read error" );
    } else {
      //Process the events
      int i = 0;
      while (i < len) {
        struct inotify_event *pevent = (struct inotify_event *)&buf[i];

        assert( pevent->len && "Assumed file name" );
        string concernedFile = watchedFolders[ pevent->wd ] + string( "/" ) +
            string( pevent->name );
        DOUT( concernedFile << " was " );
        if (pevent->mask & IN_CREATE) {
          DOUT( " created" );
          handlePlaylistUpdate( concernedFile );
        }
        if (pevent->mask & IN_MODIFY) {
          // Buffer modifications for that we do not do multiple playlist
          // updates when we receive a series of modification events - a
          // single update is sufficient then, too
          filesToUpdate.insert( concernedFile );
        }
        if (pevent->mask & IN_DELETE) {
          DOUT( " deleted" );
          handlePlaylistRemoval( concernedFile );
        }
        DOUT( endl );

        i += sizeof(struct inotify_event) + pevent->len;
      }
    }

    // Process buffered updates
    for ( set<string>::const_iterator itToUpdate = filesToUpdate.begin();
          itToUpdate != filesToUpdate.end(); itToUpdate++ ) {
      DOUT( *itToUpdate << " was modified" << endl );
      handlePlaylistUpdate( *itToUpdate );
    }
  }

  return 0;
}
